#include "Main.h"

void MainApp::resetCamera(){
	position = vec3(0, 170, -210);
	wx = 0.6f;
	wy = 0;
	wz = 0;
}

bool MainApp::init(){
	candle = new OpenGLModel();
	candle->loadFromFile("../Models/Candle.hmdl");

	flame = new ParticleSystem();
	flame->setSpawnRate(20);
	flame->setSpeed(35, 0);
	flame->setLife(0.5f, 0);
	flame->setDirectionalForce(vec3(0, 25, 0));
	flame->setFrictionFactor(1.0f);
	flame->setPosition(vec3(0, 0, 0));
	flame->setSize(9, 0);


	float k = 0.3f;
	flame->setColor(0,  k * vec4(0.0f, 0.0f, 0.0f, 0));
	flame->setColor(1,  k * vec4(1.0f, 0.4f, 0.0f, 0));
	flame->setColor(2,  k * vec4(1.0f, 0.6f, 0.2f, 0));
	flame->setColor(3,  k * vec4(1.0f, 0.8f, 0.4f, 0));
	flame->setColor(4,  k * vec4(1.0f, 1.0f, 0.6f, 0));
	flame->setColor(5,  k * vec4(1.0f, 1.0f, 0.8f, 0));
	flame->setColor(6,  k * vec4(0.9f, 1.0f, 1.0f, 0));
	flame->setColor(7,  k * vec4(0.7f, 0.8f, 1.0f, 0));
	flame->setColor(8,  k * vec4(0.5f, 0.6f, 1.0f, 0));
	flame->setColor(9,  k * vec4(0.3f, 0.4f, 1.0f, 0));
	flame->setColor(10, k * vec4(0.1f, 0.2f, 1.0f, 0));
	flame->setColor(11, k * vec4(0.0f, 0.0f, 0.0f, 0));

	return true;
}

bool MainApp::exit(){
	delete flame;

	delete candle;
	return true;
}

bool MainApp::load(){
	if (!GL_ARB_shader_objects_supported || !GL_ARB_vertex_shader_supported || !GL_ARB_fragment_shader_supported || !GL_ARB_shading_language_100_supported){
		addToLog("No GLSL support (GL_ARB_shader_objects, GL_ARB_vertex_shader, GL_ARB_fragment_shader, GL_ARB_shading_language_100)\n");
		return false;
	}
	if (!GL_EXT_blend_color_supported){
		addToLog("No constant color blend support (GL_EXT_blend_color)\n");
		return false;
	}

	if ((blur0 = renderer->addRenderTexture(256, 256, FORMAT_RGBA8, false)) == TEXTURE_NONE) return false;
	if ((blur1 = renderer->addRenderTexture(256, 256, FORMAT_RGBA8, false)) == TEXTURE_NONE) return false;

	if ((lighting = renderer->addShader("lighting.shd")) == SHADER_NONE) return false;
	if ((gauss0 = renderer->addShader("GaussianBlur.shd", "#define INVSIZE vec2(1.0 / 256.0, 0.0)\n#define k 1.0\n#define bias 0.0\n")) == SHADER_NONE) return false;
	if ((gauss1 = renderer->addShader("GaussianBlur.shd", "#define INVSIZE vec2(0.0, 1.0 / 256.0)\n#define k 0.995\n#define bias 0.002\n")) == SHADER_NONE) return false;

	setDefaultFont("../Textures/Fonts/Future.font", "../Textures/Fonts/Future.dds");

	if ((particle = renderer->addTexture("../Textures/Particle.png")) == TEXTURE_NONE) return false;

	candle->uploadToVertexBuffer();

	renderer->changeRenderTarget(blur0, TEXTURE_NONE);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	renderer->changeRenderTarget(blur1, TEXTURE_NONE);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	renderer->changeToDefaultRenderTarget();
	
	return true;
}

bool MainApp::unload(){
	return true;
}

void MainApp::drawCandles(bool lit){
	if (lit){
		renderer->setShader(lighting);
		renderer->apply();

		renderer->changeShaderConstant3f("camPos", position);
		for (unsigned int i = 0; i < sizeof(candlePos) / sizeof(vec3); i++){
			renderer->changeShaderConstant3f("lightPos", candlePos[i] + vec3(0, 70, 0));

			if (i == 1) renderer->changeBlending(ONE, ONE);
			for (unsigned int j = 0; j < sizeof(candlePos) / sizeof(vec3); j++){
				renderer->changeShaderConstant3f("candlePos", candlePos[j]);

				candle->draw();
			}
		}
	} else {
		renderer->setMask(DEPTH);
		renderer->apply();

		for (unsigned int i = 0; i < sizeof(candlePos) / sizeof(vec3); i++){
			glPushMatrix();
				glTranslatef(candlePos[i].x, candlePos[i].y, candlePos[i].z);
				glScalef(10, 10, 10);
				candle->draw();
			glPopMatrix();
		}
	}
}

void MainApp::drawFlames(){
	glClientActiveTextureARB(GL_TEXTURE0_ARB);
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glEnableClientState(GL_COLOR_ARRAY);

	vec3 dx(modelView.elem[0][0], modelView.elem[0][1], modelView.elem[0][2]);
	vec3 dy(modelView.elem[1][0], modelView.elem[1][1], modelView.elem[1][2]);
	vec3 dz(modelView.elem[2][0], modelView.elem[2][1], modelView.elem[2][2]);

	char *vertices = flame->getVertexArray(dx, dy, true, false);

	renderer->setTextures(particle);
	renderer->setBlending(ONE, ONE);
	renderer->setMask(COLOR);
	renderer->apply();

	glVertexPointer  (3, GL_FLOAT, sizeof(vec3) + sizeof(vec2) + sizeof(vec4), vertices);
	glTexCoordPointer(2, GL_FLOAT, sizeof(vec3) + sizeof(vec2) + sizeof(vec4), vertices + sizeof(vec3));
	glColorPointer   (4, GL_FLOAT, sizeof(vec3) + sizeof(vec2) + sizeof(vec4), vertices + sizeof(vec3) + sizeof(vec2));

	for (unsigned int i = 0; i < sizeof(candlePos) / sizeof(vec3); i++){
		glPushMatrix();
		vec3 pos = candlePos[i] + vec3(0, 51, 0);
		glTranslatef(pos.x, pos.y, pos.z);

		glDrawArrays(GL_QUADS, 0, flame->getParticleCount() * 4);

		glPopMatrix();
	}

	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glDisableClientState(GL_COLOR_ARRAY);
}

void MainApp::drawQuad(){
	glBegin(GL_QUADS);
		glTexCoord2f(0, 0);
		glVertex2f(-1, -1);
		glTexCoord2f(1, 0);
		glVertex2f(1, -1);
		glTexCoord2f(1, 1);
		glVertex2f(1, 1);
		glTexCoord2f(0, 1);
		glVertex2f(-1, 1);
	glEnd();
}

bool MainApp::drawFrame(){
	flame->update(time);

	projection = projectionMatrixX(1.5f, float(height) / float(width), 1, 2000);
	modelView = rotateZXY(-wx, -wy, -wz);
	modelView.translate(-position);

	static int c0, c1;
	static float lastTime = 1;
	float fTime = fmodf(time, 0.7f) / 0.7f;

	if (fTime < lastTime){
		c0 = rand() & 3;
		c1 = rand() & 3;
	}
	lastTime = fTime;

	candlePos[0] = vec3(-150, 0, 0);
	candlePos[1] = vec3( -50, 0, 0);
	candlePos[2] = vec3(  50, 0, 0);
	candlePos[3] = vec3( 150, 0, 0);

	vec3 mid = 0.5f * (candlePos[c0] + candlePos[c1]); 
	float r = length(mid - candlePos[c0]);
	float ang = fTime * PI;

	candlePos[c0] = mid + vec3( r * cosf(ang), 50 * sinf(ang),  r * sinf(ang));
	candlePos[c1] = mid + vec3(-r * cosf(ang), 50 * sinf(ang), -r * sinf(ang));



	renderer->changeRenderTarget(blur0, TEXTURE_NONE);
		glClear(GL_DEPTH_BUFFER_BIT);

		glMatrixMode(GL_PROJECTION);
		glLoadMatrixf(transpose(projection));

		glMatrixMode(GL_MODELVIEW);
		glLoadMatrixf(transpose(modelView));

		drawCandles(false);
		drawFlames();

	renderer->changeRenderTarget(blur1, TEXTURE_NONE);
		renderer->setShader(gauss0);
		renderer->setTexture("Texture", blur0);
		renderer->setDepthFunc(DEPTH_NONE);
		renderer->apply();

		drawQuad();

	renderer->changeRenderTarget(blur0, TEXTURE_NONE);
		renderer->setShader(gauss1);
		renderer->setTexture("Texture", blur1);
		renderer->setDepthFunc(DEPTH_NONE);
		renderer->apply();

		drawQuad();

	renderer->changeToDefaultRenderTarget();

	
	renderer->changeMask(ALL);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);


	glMatrixMode(GL_PROJECTION);
	glLoadMatrixf(transpose(projection));

    glMatrixMode(GL_MODELVIEW);
	glLoadMatrixf(transpose(modelView));

	glEnable(GL_CULL_FACE);

	drawCandles(true);
	drawFlames();

	glDisable(GL_CULL_FACE);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

    glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	renderer->setTextures(blur0);
	renderer->setDepthFunc(DEPTH_NONE);
	renderer->setBlending(GL_CONSTANT_COLOR_EXT, GL_ONE);
	renderer->apply();

	glBlendColorEXT(1, 0.5f, 0, 0);

	glColor3f(1, 1, 1);
	glBegin(GL_QUADS);
		glTexCoord2f(0, 0);
		glVertex2f(-1, -1);
		glTexCoord2f(1, 0);
		glVertex2f(1, -1);
		glTexCoord2f(1, 1);
		glVertex2f(1, 1);
		glTexCoord2f(0, 1);
		glVertex2f(-1, 1);
	glEnd();

	return true;
}

App *app = new MainApp();
